# -*- coding: utf-8 -*-
# Copyright (C) 2020-2021  Nexedi SA and Contributors.
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

# verify general functionality

import subprocess
import sys
import re
import signal
import time
from os.path import dirname, exists
from golang import chan, select, default, func, defer
from golang import context, sync
import psutil

import pytest

import nxdtest


@pytest.fixture
def run_nxdtest(tmpdir):
    """Fixture which returns a function which invokes nxdtest in a temporary
    directory, with the provided .nxdtest file content and with arguments
    passed as `argv`.
    """

    def _run_nxdtest(nxdtest_file_content, argv=("nxdtest",)):
        with tmpdir.as_cwd():
            with open(".nxdtest", "w") as f:
                f.write(nxdtest_file_content)
            sys_argv = sys.argv
            sys.argv = argv
            try:
                nxdtest.main()
            finally:
                sys.argv = sys_argv

    return _run_nxdtest


def test_main(run_nxdtest, capsys):
    run_nxdtest(
        """\
TestCase('TESTNAME', ['echo', 'TEST OUPUT'])
"""
    )

    captured = capsys.readouterr()
    output_lines = captured.out.splitlines()
    assert ">>> TESTNAME" in output_lines
    assert "$ echo TEST OUPUT" in output_lines
    assert "TEST OUPUT" in output_lines
    assert re.match("ok\tTESTNAME\t.*s\t# 1t 0e 0f 0s", output_lines[-2])
    assert re.match(u"# ran 1 test case:  1·ok", output_lines[-1])


def test_error_invoking_command(run_nxdtest, capsys):
    run_nxdtest(
        """\
TestCase('TESTNAME', ['not exist command'])
"""
    )

    captured = capsys.readouterr()
    assert "No such file or directory" in captured.err


def test_error_invoking_summary(run_nxdtest, capsys):
    run_nxdtest(
        """\
TestCase('TESTNAME', ['echo'], summaryf="error")
"""
    )

    captured = capsys.readouterr()
    assert "TypeError" in captured.err


def test_error_execing_nxdtest_file(run_nxdtest, capsys):
    with pytest.raises(ZeroDivisionError) as excinfo:
        run_nxdtest(
        """\
1 / 0
"""
    )
    assert '1 / 0' in str(excinfo.traceback[-1])
    # The actual .nxdtest filename is also included in traceback
    assert ".nxdtest':1" in str(excinfo.traceback[-1])


def test_run_argument(run_nxdtest, capsys):
    run_nxdtest(
        """\
TestCase('TEST1', ['echo', 'TEST1'])
TestCase('TEST2', ['echo', 'TEST2'])
""",
        argv=["nxdtest", "--run", "TEST1"],
    )
    captured = capsys.readouterr()
    assert "TEST1" in captured.out
    assert "TEST2" not in captured.out

    # --run uses search
    run_nxdtest(
        """\
TestCase('TEST1', ['echo', 'TEST1'])
TestCase('TEST2', ['echo', 'TEST2'])
TestCase('TEST10', ['echo', 'TEST10'])
""",
        argv=["nxdtest", "--run", "ST1"],
    )
    captured = capsys.readouterr()
    assert "TEST1" in captured.out
    assert "TEST10" in captured.out
    assert "TEST2" not in captured.out


# verify that nxdtest detects leaked processes.
@pytest.mark.timeout(timeout=10)
def test_run_procleak(run_nxdtest, capsys):
    procleak = "%s/testprog/procleak" % (dirname(__file__),)

    # run nxdtest in thread so that timeout handling works
    # ( if nxdtest is run on main thread, then non-py wait in WorkGroup.wait, if
    #   stuck, prevents signals from being handled at python-level )
    wg = sync.WorkGroup(context.background())
    done = chan()
    @func
    def _(ctx):
        defer(done.close)

        run_nxdtest(
            """\
TestCase('TEST_WITH_PROCLEAK', ['%s', 'AAA', 'BBB', 'CCC'])
"""     % procleak
        )
    wg.go(_)

    while 1:
        _, _rx = select(
                    default,    # 0
                    done.recv,  # 1
        )
        if _ == 0:
            time.sleep(0.1)
            continue
        wg.wait()
        break

    captured = capsys.readouterr()
    assert "AAA: terminating" in captured.out
    assert "BBB: terminating" in captured.out
    assert "CCC: terminating" in captured.out


@pytest.fixture
def distributor_with_cancelled_test(mocker):
    """A distributor for a test result with one test result line named TEST1.

    test_result.isAlive() will return False after 2 invocations, to simulate
    a test_result that was cancelled by distributor.
    """
    def _retryRPC(func_id, args=()):
        if func_id == 'getProtocolRevision':
            return 1
        assert False, ('unexpected RPC call', (func_id, args))
    mocker.patch(
        'erp5.util.taskdistribution.RPCRetry._retryRPC',
        side_effect=_retryRPC)

    test_result_line_proxy = mocker.patch(
        'erp5.util.taskdistribution.TestResultLineProxy',
        autospec=True)
    type(test_result_line_proxy).name = mocker.PropertyMock(return_value='TEST1')

    test_result_proxy = mocker.patch(
        'erp5.util.taskdistribution.TestResultProxy',
        autospec=True)
    test_result_proxy.start.side_effect = [test_result_line_proxy, None]
    test_result_proxy.isAlive.side_effect = [True, True, False]

    mocked_createTestResult = mocker.patch(
        'erp5.util.taskdistribution.TaskDistributor.createTestResult',
        return_value=test_result_proxy)

    yield 

    mocked_createTestResult.assert_called_once()
    test_result_proxy.start.assert_called()
    test_result_proxy.isAlive.assert_called()
    test_result_line_proxy.stop.assert_called()


@pytest.mark.timeout(timeout=10)
def test_distributor_timeout(run_nxdtest, capsys, tmp_path, distributor_with_cancelled_test, mocker):
    # nxdtest polls every 5 minutes, but in test we don't want to wait so long
    mocker.patch('nxdtest.time.minute', 0.05)

    slow = "%s/testprog/slow" % (dirname(__file__),)
    pidfile = str(tmp_path / 'slow.pid')
    run_nxdtest(
        """\
TestCase('TEST1', ['%s', '%s'])
""" % (slow, pidfile),
        argv=[
            "nxdtest",
            "--master_url", "http://localhost",
        ],
    )
    captured = capsys.readouterr()
    assert "TEST1" in captured.out
    assert "# test run canceled" in captured.out
    assert captured.err == ''

    with open(pidfile) as f:
        pid = int(f.read())
    assert not psutil.pid_exists(pid)


@pytest.mark.timeout(timeout=10)
@pytest.mark.parametrize('signal_number', [signal.SIGINT, signal.SIGTERM])
def test_interrupted_test(run_nxdtest, capsys, tmp_path, tmpdir, signal_number):
    slow = "%s/testprog/slow" % (dirname(__file__),)
    pidfile = str(tmp_path / 'slow.pid')

    with tmpdir.as_cwd():
        with open('.nxdtest', 'w') as f:
            f.write(
        """\
TestCase('TEST1', ['%s', '%s'])
""" % (slow, pidfile))
        nxdtest_process = subprocess.Popen([sys.executable, '%s/__init__.py' % dirname(nxdtest.__file__)])
        # wait for process to start
        for _ in range(10):
            if exists(pidfile):
                break
            time.sleep(.5)
        nxdtest_process.send_signal(signal_number)
        nxdtest_process.wait()

    with open(pidfile) as f:
        pid = int(f.read())
    assert not psutil.pid_exists(pid)
